@@ -72,7 +72,7 @@ static NTSTATUS spoolss__op_ndr_pull(struct dcesrv_call_state *dce_call, TALLOC_
 	return NT_STATUS_OK;
 }
 
-/* Note that received_packets are allocated in talloc_autofree_context(),
+/* Note that received_packets are allocated on the NULL context
  * because no other context appears to stay around long enough. */
 static struct received_packet {
 	uint16_t opnum;
@@ -80,6 +80,20 @@ static struct received_packet {
 	struct received_packet *prev, *next;
 } *received_packets = NULL;
 
+static void free_received_packets(void)
+{
+	struct received_packet *rp;
+	struct received_packet *rp_next;
+
+	for (rp = received_packets; rp; rp = rp_next) {
+		rp_next = rp->next;
+		DLIST_REMOVE(received_packets, rp);
+		talloc_unlink(rp, rp->r);
+		talloc_free(rp);
+	}
+	received_packets = NULL;
+}
+
 static WERROR _spoolss_ReplyOpenPrinter(struct dcesrv_call_state *dce_call,
 					TALLOC_CTX *mem_ctx,
 					struct spoolss_ReplyOpenPrinter *r)
@@ -136,11 +150,11 @@ static NTSTATUS spoolss__op_dispatch(struct dcesrv_call_state *dce_call, TALLOC_
 	uint16_t opnum = dce_call->pkt.u.request.opnum;
 	struct received_packet *rp;
 
-	rp = talloc_zero(talloc_autofree_context(), struct received_packet);
+	rp = talloc_zero(NULL, struct received_packet);
 	rp->opnum = opnum;
 	rp->r = talloc_reference(rp, r);
 
-	DLIST_ADD_END(received_packets, rp, struct received_packet *);
+	DLIST_ADD_END(received_packets, rp);
 
 	switch (opnum) {
 	case 58: {
@@ -247,20 +261,13 @@ static NTSTATUS spoolss__op_init_server(struct dcesrv_context *dce_ctx, const st
 static bool test_OpenPrinter(struct torture_context *tctx,
 			     struct dcerpc_pipe *p,
 			     struct policy_handle *handle,
-			     const char *name)
+			     const char *printername)
 {
 	struct spoolss_OpenPrinter r;
-	const char *printername;
 	struct dcerpc_binding_handle *b = p->binding_handle;
 
 	ZERO_STRUCT(r);
 
-	if (name) {
-		printername	= talloc_asprintf(tctx, "\\\\%s\\%s", dcerpc_server_name(p), name);
-	} else {
-		printername	= talloc_asprintf(tctx, "\\\\%s", dcerpc_server_name(p));
-	}
-
 	r.in.printername	= printername;
 	r.in.datatype		= NULL;
 	r.in.devmode_ctr.devmode= NULL;
@@ -302,6 +309,7 @@ static struct spoolss_NotifyOption *setup_printserver_NotifyOption(struct tortur
 	return o;
 }
 
+#if 0
 static struct spoolss_NotifyOption *setup_printer_NotifyOption(struct torture_context *tctx)
 {
 	struct spoolss_NotifyOption *o;
@@ -321,7 +329,7 @@ static struct spoolss_NotifyOption *setup_printer_NotifyOption(struct torture_co
 
 	return o;
 }
-
+#endif
 
 static bool test_RemoteFindFirstPrinterChangeNotifyEx(struct torture_context *tctx,
 						      struct dcerpc_binding_handle *b,
@@ -352,17 +360,17 @@ static bool test_RemoteFindFirstPrinterChangeNotifyEx(struct torture_context *tc
 static bool test_RouterRefreshPrinterChangeNotify(struct torture_context *tctx,
 						  struct dcerpc_binding_handle *b,
 						  struct policy_handle *handle,
-						  struct spoolss_NotifyOption *options)
+						  struct spoolss_NotifyOption *options,
+						  struct spoolss_NotifyInfo **info)
 {
 	struct spoolss_RouterRefreshPrinterChangeNotify r;
-	struct spoolss_NotifyInfo *info;
 
 	torture_comment(tctx, "Testing RouterRefreshPrinterChangeNotify\n");
 
 	r.in.handle = handle;
 	r.in.change_low = 0;
 	r.in.options = options;
-	r.out.info = &info;
+	r.out.info = info;
 
 	torture_assert_ntstatus_ok(tctx, dcerpc_spoolss_RouterRefreshPrinterChangeNotify_r(b, tctx, &r),
 		"RouterRefreshPrinterChangeNotify failed");
@@ -372,6 +380,7 @@ static bool test_RouterRefreshPrinterChangeNotify(struct torture_context *tctx,
 	return true;
 }
 
+#if 0
 static bool test_SetPrinter(struct torture_context *tctx,
 			    struct dcerpc_pipe *p,
 			    struct policy_handle *handle)
@@ -425,6 +434,7 @@ static bool test_SetPrinter(struct torture_context *tctx,
 
 	return true;
 }
+#endif
 
 static bool test_start_dcerpc_server(struct torture_context *tctx,
 				     struct tevent_context *event_ctx,
@@ -455,8 +465,8 @@ static bool test_start_dcerpc_server(struct torture_context *tctx,
 
 	lpcfg_set_cmdline(tctx->lp_ctx, "dcerpc endpoint servers", "spoolss");
 
-	load_interfaces(tctx, lpcfg_interfaces(tctx->lp_ctx), &ifaces);
-	address = iface_n_ip(ifaces, 0);
+	load_interface_list(tctx, tctx->lp_ctx, &ifaces);
+	address = iface_list_first_v4(ifaces);
 
 	torture_comment(tctx, "Listening for callbacks on %s\n", address);
 
@@ -464,7 +474,9 @@ static bool test_start_dcerpc_server(struct torture_context *tctx,
 	torture_assert_ntstatus_ok(tctx, status,
 				   "unable to initialize process models");
 
-	status = smbsrv_add_socket(tctx, event_ctx, tctx->lp_ctx, process_model_startup("single"), address);
+	status = smbsrv_add_socket(tctx, event_ctx, tctx->lp_ctx,
+				   process_model_startup("single"),
+				   address, NULL);
 	torture_assert_ntstatus_ok(tctx, status, "starting smb server");
 
 	status = dcesrv_init_context(tctx, tctx->lp_ctx, endpoints, &dce_ctx);
@@ -473,7 +485,8 @@ static bool test_start_dcerpc_server(struct torture_context *tctx,
 
 	for (e=dce_ctx->endpoint_list;e;e=e->next) {
 		status = dcesrv_add_ep(dce_ctx, tctx->lp_ctx,
-				       e, tctx->ev, process_model_startup("single"));
+				       e, tctx->ev,
+				       process_model_startup("single"), NULL);
 		torture_assert_ntstatus_ok(tctx, status,
 				"unable listen on dcerpc endpoint server");
 	}
@@ -487,7 +500,8 @@ static bool test_start_dcerpc_server(struct torture_context *tctx,
 static struct received_packet *last_packet(struct received_packet *p)
 {
 	struct received_packet *tmp;
-	for (tmp = p; tmp->next; tmp = tmp->next) ;;
+	for (tmp = p; tmp->next; tmp = tmp->next) {
+	}
 	return tmp;
 }
 
@@ -503,31 +517,37 @@ static bool test_RFFPCNEx(struct torture_context *tctx,
 	struct spoolss_NotifyOption *printer_option = setup_printer_NotifyOption(tctx);
 #endif
 	struct dcerpc_binding_handle *b = p->binding_handle;
+	const char *printername = NULL;
+	struct spoolss_NotifyInfo *info = NULL;
 
-	received_packets = NULL;
+	free_received_packets();
 
 	/* Start DCE/RPC server */
-	torture_assert(tctx, test_start_dcerpc_server(tctx, p->conn->event_ctx, &dce_ctx, &address), "");
+	torture_assert(tctx, test_start_dcerpc_server(tctx, tctx->ev, &dce_ctx, &address), "");
 
-	torture_assert(tctx, test_OpenPrinter(tctx, p, &handle, NULL), "");
+	printername	= talloc_asprintf(tctx, "\\\\%s", dcerpc_server_name(p));
+
+	torture_assert(tctx, test_OpenPrinter(tctx, p, &handle, printername), "");
 	torture_assert(tctx, test_RemoteFindFirstPrinterChangeNotifyEx(tctx, b, &handle, address, server_option), "");
 	torture_assert(tctx, received_packets, "no packets received");
 	torture_assert_int_equal(tctx, received_packets->opnum, NDR_SPOOLSS_REPLYOPENPRINTER,
 		"no ReplyOpenPrinter packet after RemoteFindFirstPrinterChangeNotifyEx");
-	torture_assert(tctx, test_RouterRefreshPrinterChangeNotify(tctx, b, &handle, NULL), "");
-	torture_assert(tctx, test_RouterRefreshPrinterChangeNotify(tctx, b, &handle, server_option), "");
+	torture_assert(tctx, test_RouterRefreshPrinterChangeNotify(tctx, b, &handle, NULL, &info), "");
+	torture_assert(tctx, test_RouterRefreshPrinterChangeNotify(tctx, b, &handle, server_option, &info), "");
 	torture_assert(tctx, test_ClosePrinter(tctx, b, &handle), "");
 	tmp = last_packet(received_packets);
 	torture_assert_int_equal(tctx, tmp->opnum, NDR_SPOOLSS_REPLYCLOSEPRINTER,
 		"no ReplyClosePrinter packet after ClosePrinter");
 #if 0
+	printername	= talloc_asprintf(tctx, "\\\\%s\\%s", dcerpc_server_name(p), name);
+
 	torture_assert(tctx, test_OpenPrinter(tctx, p, &handle, "Epson AL-2600"), "");
 	torture_assert(tctx, test_RemoteFindFirstPrinterChangeNotifyEx(tctx, p, &handle, address, printer_option), "");
 	tmp = last_packet(received_packets);
 	torture_assert_int_equal(tctx, tmp->opnum, NDR_SPOOLSS_REPLYOPENPRINTER,
 		"no ReplyOpenPrinter packet after RemoteFindFirstPrinterChangeNotifyEx");
-	torture_assert(tctx, test_RouterRefreshPrinterChangeNotify(tctx, p, &handle, NULL), "");
-	torture_assert(tctx, test_RouterRefreshPrinterChangeNotify(tctx, p, &handle, printer_option), "");
+	torture_assert(tctx, test_RouterRefreshPrinterChangeNotify(tctx, p, &handle, NULL, &info), "");
+	torture_assert(tctx, test_RouterRefreshPrinterChangeNotify(tctx, p, &handle, printer_option, &info), "");
 	torture_assert(tctx, test_SetPrinter(tctx, p, &handle), "");
 	tmp = last_packet(received_packets);
 	torture_assert_int_equal(tctx, tmp->opnum, NDR_SPOOLSS_ROUTERREPLYPRINTEREX,
@@ -539,6 +559,7 @@ static bool test_RFFPCNEx(struct torture_context *tctx,
 #endif
 	/* Shut down DCE/RPC server */
 	talloc_free(dce_ctx);
+	free_received_packets();
 
 	return true;
 }
